
/*
 * File ArrayTheoryISE.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-25. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-25 only because so is required by the CASC-25
 * rules, and for no other reason.
 */


#include "Inferences/ArrayTheoryISE.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Theory.hpp"
#include "Lib/DArray.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Random.hpp"
#include "Shell/Skolem.hpp"
#include "Shell/Statistics.hpp"

namespace Inferences
{

using namespace Lib;
using namespace Kernel;
using namespace Inferences;

ArrayTermTransformer::ArrayTermTransformer() :
  _select1Functor(env.signature->getInterpretingSymbol(Theory::SELECT1_INT)),
  _select2Functor(env.signature->getInterpretingSymbol(Theory::SELECT2_INT)),
  _store1Functor(env.signature->getInterpretingSymbol(Theory::STORE1_INT)),
  _store2Functor(env.signature->getInterpretingSymbol(Theory::STORE2_INT))
{}

ArrayTermTransformer::~ArrayTermTransformer() {}

TermList ArrayTermTransformer::transformSubterm(TermList trm)
{
  CALL("ArrayTermTransformer::transformSubterm");
  
  if (!trm.isTerm()) {
    return trm;
  }
  
  Term* term = trm.term();
  
  if (term->functor() == _select1Functor ||
      term->functor() == _select2Functor)
  {
    TermList* array = term->nthArgument(0);
    
    if (array->isTerm() &&
        (array->term()->functor() == _store1Functor ||
         array->term()->functor() == _store2Functor))
    {
      
      Term* store = term->nthArgument(0)->term();
      
      TermList* selectIndex = term->nthArgument(1);
      TermList* storeIndex = store->nthArgument(1);
      
      if (selectIndex->sameContent(storeIndex)) {
        return *store->nthArgument(2);
      }
    }
  } else if (term->functor() == _store1Functor ||
             term->functor() == _store2Functor)
  {
    TermList* array = term->nthArgument(0);
    TermList* value = term->nthArgument(2);
    
    if (array->isTerm() &&
        (array->term()->functor() == _store1Functor ||
         array->term()->functor() == _store2Functor))
    {
      
      Term* store = array->term();
      TermList* outerIndex = term->nthArgument(1);
      TermList* innerIndex = store->nthArgument(1);
      
      if (outerIndex->sameContent(innerIndex)) {
        TermList args[3];
        args[0] = *store->nthArgument(0);
        args[1] = *outerIndex;
        args[2] = *value;
        return TermList(Term::create(term, args));
      }
    } else if (value->isTerm() &&
               (value->term()->functor() == _select1Functor ||
                value->term()->functor() == _select2Functor))
    {
      
      TermList* selectArray = value->term()->nthArgument(0);
      TermList* selectIndex = value->term()->nthArgument(1);
      TermList* storeIndex = term->nthArgument(1);
      
      if (selectArray->sameContent(array) &&
          selectIndex->sameContent(storeIndex)) {
        return *array;
      }
    }
  }
  
  return trm;
}

Literal* ArrayTermTransformer::rewriteNegEqByExtensionality(Literal* l)
{
  CALL("ArrayTermTransformer::rewriteNegEqByExtensionality");
  
  if (l->isEquality() && l->isNegative()) {
    unsigned sort = SortHelper::getEqualityArgumentSort(l);
    unsigned select;
    unsigned valSort;

    if (sort == Sorts::SRT_ARRAY1) {
      select = _select1Functor;
      valSort = Sorts::SRT_INTEGER;
    } else if (sort == Sorts::SRT_ARRAY2) {
      select = _select2Functor;
      valSort = Sorts::SRT_ARRAY1;
    } else {
      return l;
    }
    
    
    TermList* lhs = l->nthArgument(0);
    TermList* rhs = l->nthArgument(1);
    
    
    
    TermList params[] = {*lhs, *rhs};
    unsigned sk = theory->getArrayExtSkolemFunction(sort);
    TermList skolemTerm(Term::create(sk, 2, params));
    
    TermList sel1(Term::create2(select, *lhs, skolemTerm));
    TermList sel2(Term::create2(select, *rhs, skolemTerm));
    
    return Literal::createEquality(false, sel1, sel2, valSort);
  }
  
  return l;
}

ArrayTheoryISE::ArrayTheoryISE() :
  _transformer(ArrayTermTransformer()) {}

Clause* ArrayTheoryISE::simplify(Clause* c)
{
  CALL("ArrayTheoryISE::simplify");
  
  if (c->inputType() == Unit::AXIOM) {
    return c;
  }
  
  static DArray<Literal*> lits(32);
  int length = c->length();
  lits.ensure(length);
  bool modified = false;
  
  
  
  for (int i = 0; i < length; ++i) {
    Literal* l = (*c)[i];
    bool done = false;
    Literal* transformed;
    
    
    
    
    while (!done) {
      
      transformed = _transformer.transform(l);
      
      
      if (l != transformed) {
        modified = true;
        l = transformed;
      } else {
        done = true;
      }
    }
    
    Literal* transformed2 = _transformer.rewriteNegEqByExtensionality(transformed);
    if (transformed2 != transformed) {
      modified = true;
    }
    
    lits[i] = transformed2;
  }
  
  if (modified) {
    
    Clause* d = new(length)
      Clause(length,
             c->inputType(),
             new Inference1(Inference::INTERPRETED_SIMPLIFICATION, c));
    for (int i = 0; i < length; ++i) {
      (*d)[i] = lits[i];
    }
    d->setAge(c->age());
    
    
    
    return d;
  }
  
  return c;
}

}
